from abc import ABC, abstractmethod
from typing import List

from .route_info import RouteInfo


class AbstractSocketRouter(ABC):
    _routes: List[RouteInfo] = []

    @property
    def routes(self) -> List[RouteInfo]:
        return self._routes

    """
    this method can simply be overridden by passing an array of RouteInfo objects like this:
    @routes.setter
    def routes(self, routes: List[RouteInfo]):
        self._routes = [
            {route: 'sayHello'}, # hydrate and dehydrate functions are optional
            {route: 'getArticles', dehydrate: multi_article_serializer}
        ]

    though this approach is simple, I highly recommend you to put this list in a `SOME.env` file, so it can be shared by both Django and React-based frontend.
    """

    @routes.setter
    @abstractmethod
    def routes(self, routes: List[RouteInfo]):
        pass

    def _get_route(self, route: str) -> RouteInfo | None:
        for route_info in self.routes:
            if route_info['route'] == route:
                return route_info
        return None



class StatusCodes:
    # Successful responses
    OK = 200
    CREATED = 201
    ACCEPTED = 202
    NO_CONTENT = 204

    # Client error responses
    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    PENDING = 402
    FORBIDDEN = 403
    NOT_FOUND = 404
    METHOD_NOT_ALLOWED = 405
    NOT_ACCEPTABLE = 406
    REQUEST_TIMEOUT = 408
    CONFLICT = 409
    GONE = 410
    LENGTH_REQUIRED = 411
    PRECONDITION_FAILED = 412
    PAYLOAD_TOO_LARGE = 413
    URI_TOO_LONG = 414
    UNSUPPORTED_MEDIA_TYPE = 415
    RANGE_NOT_SATISFIABLE = 416
    EXPECTATION_FAILED = 417
    TOO_MANY_REQUESTS = 429

    # Server error responses
    INTERNAL_SERVER_ERROR = 500
    NOT_IMPLEMENTED = 501
    BAD_GATEWAY = 502
    SERVICE_UNAVAILABLE = 503
    GATEWAY_TIMEOUT = 504
    HTTP_VERSION_NOT_SUPPORTED = 505



from typing import TypedDict, Callable, Optional, Any


class RouteInfo(TypedDict, total=False):
    """
    if the incoming message has route attribute equal to `loadNode` then the on_load_node method will be called automatically.
    no more need to parse the message to recognize what has to be done.
    """
    route: str
    """
    any function that takes headers and payload, and loads the data from it.
    deserializing can be considered as good candidate.
    another example is to fetch the entity_id from headers and load the entity using it.
    """
    hydrate: Optional[Callable[[dict[str, str],Any], Any]]
    """
    the duty of dehydrate function is to get an object and serialize it as string or dict (no complex object included).
    PS. the dehydrate function runs only if status code is successful (200 series)
    """
    dehydrate: Optional[Callable]


import re


def route_to_method_name(route: str) -> str:
    return 'on_{}'.format(re.sub(r'([a-z])([A-Z])', r'\1_\2', route).lower())

def result_is_successful(status: int) -> bool:
    return 200 <= status < 300



from django.conf import settings

from channels.generic.websocket import AsyncJsonWebsocketConsumer

from . import StatusCodes, SocketResult, BaseMessage, AbstractSocketRouter, RouteInfo
from .tools import route_to_method_name, result_is_successful

User = settings.AUTH_USER_MODEL


class AsyncSocketRouterConsumer(AsyncJsonWebsocketConsumer, AbstractSocketRouter):
    """
    To implement this class, you'll need to implement this method:

    @routes.setter
    @abstractmethod
    def routes (self, routes: List[RouteInfo]):
        pass

    This method can simply be overridden by passing an array of RouteInfo objects like this:
    @routes.setter
    def routes (self, routes: List[RouteInfo]):
        self._routes = [
            {route: 'sayHello'}, # hydrate and dehydrate functions are optional
            {route: 'getArticles', dehydrate: multi_article_serializer}
        ]

    Though this approach is straightforward, I highly recommend you to put this list in a `SOME.env` file, so it can be shared by both Django and React-based frontend.

    Then, if you have a route like `sayHello`,
    your consumer will need to have a method named on_say_hello, and there you can define the code you expect to be executed when sayHello is called.

    So when in the frontend you call sayHello route, the code in on_say_hello will be executed automatically.

    Read more in README.md about how it facilitates the entire process, using hydrate and dehydrate functions.
    """

    _user = None

    @property
    def user(self) -> User:
        return self._user

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def connect(self):
        self._user = self.scope['user']
        await self.accept()

    async def receive_json(self, content, **kwargs):
        message: BaseMessage
        try:
            message = BaseMessage.from_incoming_message(content)
            if message.route == 'PING':  # only heart-bit checks, response, so the client makes sure the connection is open
                await self.send_json(
                    BaseMessage(
                        route='PING',
                        uuid=message.uuid
                    ).serialize()
                )
                return

            # find route
            route_info: RouteInfo | None = self._get_route(message.route)
            if not route_info:
                await self.send_json(
                    BaseMessage.build_response(message, None, None, StatusCodes.NOT_FOUND).serialize()
                )
                return

            # find method
            method_name = route_to_method_name(route_info['route'])
            method = getattr(self, method_name)  # if the method doesn't exist, an exception will be raised

            # hydrate the payload if the function is provided
            payload = await route_info['hydrate'](message.headers, message.payload) if route_info.get('hydrate', False) else message.payload

            result: SocketResult = await method(payload, message.headers)

            should_dehydrate = route_info.get('dehydrate', False) and result_is_successful(result['status'])
            out_payload = await route_info['dehydrate'](result['payload']) if should_dehydrate else result['payload']

            await self.send_json(
                BaseMessage.build_response(message, result['headers'], out_payload, result['status']).serialize()
            )
        except AttributeError:
            message = BaseMessage(content.get('route'), content.get('uuid'))
            await self.send_json(
                BaseMessage.build_response(
                    message, None, {'Error': 'Route not declared properly in backend'},
                    StatusCodes.INTERNAL_SERVER_ERROR
                ).serialize()
            )
        except KeyError:
            await self.send_json(
                BaseMessage(
                    content.get('uuid', ''), content.get('route', ''), None, None, StatusCodes.BAD_REQUEST
                ).serialize()
            )
            return


from typing import Any, Self

from .status import StatusCodes


class BaseMessage:
    """
    This class provides the Abstract type of Message with required fields and methods
    that the library highly depends on them.
    """

    """
    route is used to determine which method is responsible to handle the incoming message
    the routes will convert to snakecase for making method names.
    method names will follow this: f"on_{snakecase(route)}"
    """
    _route: str
    """
    uuid acts as a tracking code, so that when server responses the client with this code,
    the client will identify which request does it belongs to. another essential part of this
    library.
    """
    _uuid: str | int
    """
    While there aren't actual headers part when using socket as transition protocol, to make it feel more like http call
    this field is added, I found it meaningful to send metadata(like API_KEY or even entity ID) through headers and use payload to send actual data 
    """
    _headers: dict | None = None
    """
    Payload field is designed to be used as main data storage. It's supposed to mostly contain a dictionary (the message
    object itself is Json coded and decoded so the rest data are not Json Coded), however any other value is possible
    like string, int, float, even a simple boolean or a huge base64 encoded string.
    """
    _payload: Any = None
    """
    Status Field is supposed to be set for sending responses, just like HTTP calls.
    """
    _status: int = StatusCodes.OK

    @property
    def route(self) -> str:
        return self._route

    @property
    def uuid(self) -> str | int:
        return self._uuid

    @property
    def headers(self) -> dict:
        return self._headers

    @property
    def payload(self):
        return self._payload

    @payload.setter
    def payload(self, value):
        self._payload = value

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status: int):
        self._status = status

    def __init__(
            self,
            route: str,
            uuid: str | int,
            headers: dict[str, str] | None = None,
            payload: Any = None,
            status: int = StatusCodes.OK
    ):
        self._route = route
        self._uuid = uuid
        self._headers = headers
        self._payload = payload
        self._status = status

    @classmethod
    def from_incoming_message(cls, incoming_message: dict):
        uuid = incoming_message.get('uuid')
        if not uuid or not isinstance(uuid, str):
            raise KeyError("uuid is missing or wrong format")
        route = incoming_message.get('route')
        if not route or not isinstance(route, str):
            raise KeyError("proper route is missing")

        headers = incoming_message.get('headers')
        headers = headers if isinstance(headers, dict) else None
        payload = incoming_message.get('payload')
        message = BaseMessage(route=route, uuid=uuid, headers=headers, payload=payload)
        return message

    @classmethod
    def build_response(cls, income_message: Self, headers: dict[str,str] | None = None,payload: Any = None, status: int = StatusCodes.OK):
        return BaseMessage(income_message.route, income_message.uuid, headers, payload, status)

    def serialize(self) -> dict:
        return {
            'uuid': self.uuid,
            'route': self.route,
            'headers': self.headers,
            'payload': self.payload,
            'status': self.status
        }